/********************************************************* begin of preamble
**
** Copyright (C) 2003-2010 Software- und Organisations-Service GmbH. 
** All rights reserved.
**
** This file may be used under the terms of either the 
**
**   GNU General Public License version 2.0 (GPL)
**
**   as published by the Free Software Foundation
**   http://www.gnu.org/licenses/gpl-2.0.txt and appearing in the file
**   LICENSE.GPL included in the packaging of this file. 
**
** or the
**  
**   Agreement for Purchase and Licensing
**
**   as offered by Software- und Organisations-Service GmbH
**   in the respective terms of supply that ship with this file.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
** IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
** THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
** BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
** SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
** CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
** POSSIBILITY OF SUCH DAMAGE.
********************************************************** end of preamble*/
package com.sos.scheduler.engine.kernel.plugin;

import com.sos.scheduler.engine.kernel.command.CommandHandler;
import com.sos.scheduler.engine.kernel.AbstractHasPlatform;
import com.sos.scheduler.engine.kernel.Scheduler;
import com.sos.scheduler.engine.kernel.SchedulerException;
import com.sos.scheduler.engine.kernel.Subsystem;
import com.sos.scheduler.engine.kernel.command.HasCommandHandlers;
import com.sos.scheduler.engine.kernel.log.PrefixLog;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import org.w3c.dom.Element;
import static com.sos.scheduler.engine.kernel.util.Util.stackTrace;
import static com.sos.scheduler.engine.kernel.util.XmlUtils.*;
import static java.util.Arrays.asList;

//TODO Eigenes PrefixLog einführen

public class PlugInSubsystem extends AbstractHasPlatform implements Subsystem, HasCommandHandlers {
    private static final String staticFactoryMethodName = "factory";
    private final Scheduler scheduler;
    private final Map<String,PluginAdapter> plugIns = new HashMap<String,PluginAdapter>();
    private final CommandHandler[] commandHandlers = { 
        new PlugInCommandExecutor(this), 
        new PlugInCommandCommandXmlParser(this),
        new PluginCommandResultXmlizer(this) };

    
    public PlugInSubsystem(Scheduler scheduler) {
        super(scheduler.getPlatform());
        this.scheduler = scheduler;
    }


    public final void load(Element root) {
        Element plugInsElement = elementXPathOrNull(root, "config/plugins");
        if (plugInsElement != null) {
            for (Element e: elementsXPath(plugInsElement, "plugin"))  tryAddPlugIn(e);
        }
    }


    private void tryAddPlugIn(Element e) {
        String className = e.getAttribute("java_class");
        if (className.isEmpty())  throw new SchedulerException("Missing attribute java_class in <plugin>");
        Element contentElement = elementXPathOrNull(e, "plugin.config");
        tryAddPlugIn(className, contentElement);
    }
    

    private void tryAddPlugIn(String className, Element elementOrNull) {
        try {
            PlugIn p = newPlugIn(className, elementOrNull);
            PluginAdapter a = p instanceof CommandPlugin? new CommandPluginAdapter((CommandPlugin)p, className, log())
                : new PluginAdapter(p, className, log());
            plugIns.put(className, a);
            log().info(a + " added");
        } catch (Exception x) {
            logError(log(), "Plug-in " + className, x);
        }
    }


    private PlugIn newPlugIn(String className, Element elementOrNull) throws Exception {
        Class<?> c = Class.forName(className);
        PlugInFactory f = (PlugInFactory)c.getMethod(staticFactoryMethodName).invoke(null);
        return f.newInstance(scheduler, elementOrNull);
    }


    public final void activate() {
        for (PluginAdapter p: plugIns.values())  p.tryActivate();
    }


    public final void close() {
        for (PluginAdapter p: plugIns.values())  p.tryClose();
    }


    final CommandPluginAdapter commandPluginByClassName(String className) {
        PluginAdapter a = pluginByClassName(className);
        if (!(a instanceof CommandPluginAdapter))
            throw new SchedulerException("Plugin is not a " + CommandPlugin.class.getSimpleName());
        return (CommandPluginAdapter)a;
    }


    private PluginAdapter pluginByClassName(String className) {
        PluginAdapter result = plugIns.get(className);
        if (result == null)
            throw new SchedulerException("Unknown plugin '" + className + "'");
        return result;
    }


    @Override public final Collection<CommandHandler> getCommandHandlers() {
        return asList(commandHandlers);
    }


    static void logError(PrefixLog log, String pluginName, Throwable t) {
        log.error(pluginName + ": " + t + "\n" + stackTrace(t));
    }
}
