/********************************************************* begin of preamble
**
** Copyright (C) 2003-2010 Software- und Organisations-Service GmbH. 
** All rights reserved.
**
** This file may be used under the terms of either the 
**
**   GNU General Public License version 2.0 (GPL)
**
**   as published by the Free Software Foundation
**   http://www.gnu.org/licenses/gpl-2.0.txt and appearing in the file
**   LICENSE.GPL included in the packaging of this file. 
**
** or the
**  
**   Agreement for Purchase and Licensing
**
**   as offered by Software- und Organisations-Service GmbH
**   in the respective terms of supply that ship with this file.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
** IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
** THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
** BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
** SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
** CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
** POSSIBILITY OF SUCH DAMAGE.
********************************************************** end of preamble*/
package com.sos.scheduler.engine.kernel.util.sync;

import com.sos.scheduler.engine.kernel.util.Time;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.SynchronousQueue;
import org.apache.log4j.Logger;


/** Zur Rendezvous-Synchronisierung zweier Threads, also zum synchronsierten Aufruf eines Entrys eines anderen Threads.
 * Der rufende Thread ruft call(ARG a) und wartet damit auf den dienenden Thread, bis dieser enter() erreicht.
 * enter() setzt erst fort, wenn ein rufender Thread im call() ist. Dann beginnt das Rendezvous.
 * Der dienende Thread bekommt a während der rufende Thread wartet, bis der dienende Thread leaveException(RESULT r) ruft.
 * r wird dem rufenden Thread zurückgegeben und beide Threads setzen unabhängig fort.
 *
 * @param <ARG> Typ zu Übergabe an den dienenden Thread
 * @param <RESULT> Rückgabe des Rendevous' an den rufenden Thread
 */
public class Rendezvous<ARG,RESULT> {
    private static final Logger logger = Logger.getLogger(Rendezvous.class);

    private volatile Thread servingThread = null;
    private volatile boolean inRendezvous = false;
    private final BlockingQueue<ARG> argumentQueue = new SynchronousQueue<ARG>();
    private final BlockingQueue<Return> returnQueue = new SynchronousQueue<Return>();
    private volatile boolean closed = false;


    public final void closeServing() {
        assertIsServingThread();
        try {
            if (argumentQueue.peek() != null)   // Greift nicht immer, dafür müsste die Abfrage synchronisiert werden.
                enter(Time.of(0));

            if (inRendezvous) {
                RendezvousServerClosedException x = new RendezvousServerClosedException();
                logger.error(x, x);
                leaveException(x);
                throw x;
            }
        }
        finally {
            servingThread = null;
            closed = true;
        }
    }

    
    public final void beginServing() {
        servingThread = Thread.currentThread();
    }


    private void assertIsNotClosed() {
        if (closed)  throw new RendezvousException("Rendezvous serving is closed");
    }


    public final RESULT call(ARG o) {
        asyncCall(o);
        return awaitResult();
    }


    /** Danach syncResult() aufrufen, bis der Aufruf nicht NULL liefert **/
    public final void asyncCall(ARG o) {
        try {
            assertIsNotClosed();
            argumentQueue.put(o);
        }
        catch (InterruptedException x) { throw new RuntimeException(x); }
    }


    public final RESULT awaitResult() {
        return awaitResult(Time.eternal);
    }

    /** Wenn awaitResult() NULL liefert, den Aufruf wiederholen! */
    public final RESULT awaitResult(Time t) {
        try {
            Return r = returnQueue.poll(t.value, t.unit);
            if (r == null)
                return null;
            else {
                if  (r.runtimeException != null)  throw r.runtimeException;
                return r.result;
            }
        }
        catch (InterruptedException x) { throw new RuntimeException(x); }
    }


    /** Am Ende immer leave aufrufen! */
    public final ARG enter() {
        return enter(Time.eternal);
    }


    /** Am Ende immer leave() aufrufen! */
    public ARG enter(Time timeout) {
        try {
            assertIsServingThread();
            if (inRendezvous)  throw new IllegalStateException("Already in rendezvous");

            ARG result = timeout == Time.eternal? argumentQueue.take() : argumentQueue.poll(timeout.value, timeout.unit);  // null, if timeout
            if (result != null)  inRendezvous = true;
            return result;
        }
        catch (InterruptedException x) { throw new RuntimeException(x); }
    }


    public void leave(RESULT o) {
        leave(new Return(o));
    }


    public final void leaveException(Throwable t) {
        RuntimeException x = t instanceof RuntimeException? (RuntimeException)t : new RuntimeException(t);
        leave(new Return(x));
    }


    private void leave(Return r) {
        try {
            if (!inRendezvous)  throw new IllegalStateException("Not in rendezvous");
            returnQueue.put(r);
            inRendezvous = false;
        }
        catch (InterruptedException x) { throw new RuntimeException(x); }
    }


    private void assertIsServingThread() {
        if (servingThread == null)  throw new RendezvousException("beginServing() has not been not called");
        if (servingThread != Thread.currentThread())  throw new RendezvousException("Method must be called in serving thread only");
    }


    public final boolean isInRendezvous() {
        return inRendezvous;
    }


    private final class Return {
        private final RESULT result;
        private final RuntimeException runtimeException;
        
        private Return(RESULT r, RuntimeException x) { result = r; runtimeException = x; }
        private Return(RESULT r) { this(r, null); }
        private Return(RuntimeException x) { this(null, x); }
    }
}
